@@ -1,5 +1,6 @@
 from jsengine.abstract import AbstractJSEngine
 from jsengine.chakra import ChakraHandle
+from jsengine.v8 import MiniRacer
 from jsengine.exceptions import *
 import jsengine.detect as _d
 import json
@@ -26,18 +27,62 @@ class InternalJSEngine(AbstractJSEngine):
             pass
 
 
+class V8JSEngine(InternalJSEngine):
+    '''Wrappered for V8 python binding PyMiniRacer.'''
+
+    def __init__(self, source=u'', init_global=False, init_del_gobjects=[]):
+        if not _d.v8_available:
+            msg = ('No supported V8 package found on current python environment!'
+                   ' Please install python package PyMiniRacer')
+            if _d.quickjs_available:
+                msg += ' or use QuickJSEngine.'
+            elif _d.chakra_available:
+                msg += ' or use ChakraJSEngine.'
+            elif _d.external_interpreter:
+                msg += ' or use ExternalJSEngine.'
+            else:
+                msg += '.'
+            raise RuntimeError(msg)
+        InternalJSEngine.__init__(self, source, init_global, init_del_gobjects)
+
+    __init__.__doc__ = AbstractJSEngine.__init__.__doc__
+
+    # Here is a scope issue now, we MUST execute all codes at once.
+    # see https://github.com/sqreen/PyMiniRacer/issues/148
+    def _append(self, code):
+        pass
+
+    def _eval(self, code):
+        return self._context.eval(self.source)
+
+    class Context(object):
+        def __init__(self):
+            pass
+
+        def eval(self, code, eval=True, raw=False):
+            ok, result = MiniRacer().eval(code, raw)
+            if ok:
+                if eval:
+                    return result
+            else:
+                raise ProgramError(str(result))
+
+
 class ChakraJSEngine(InternalJSEngine):
     '''Wrappered for system's built-in Chakra or PyChakra(ChakraCore).'''
 
     def __init__(self, source=u'', init_global=False, init_del_gobjects=[]):
         if not _d.chakra_available:
-            msg = 'No supported Chakra binary found on your system!'
+            msg = ('No supported Chakra binary found on your system!'
+                   ' Please install python package PyChakra')
             if _d.quickjs_available:
-                msg += ' Please install PyChakra or use QuickJSEngine.'
+                msg += ' or use QuickJSEngine.'
+            elif _d.v8_available:
+                msg += ' or use V8JSEngine.'
             elif _d.external_interpreter:
-                msg += ' Please install PyChakra or use ExternalJSEngine.'
+                msg += ' or use ExternalJSEngine.'
             else:
-                msg += ' Please install PyChakra.'
+                msg += '.'
             raise RuntimeError(msg)
         InternalJSEngine.__init__(self, source, init_global, init_del_gobjects)
 
@@ -61,13 +106,16 @@ class QuickJSEngine(InternalJSEngine):
 
     def __init__(self, source=u'', init_global=False, init_del_gobjects=[]):
         if not _d.quickjs_available:
-            msg = 'No supported QuickJS package found on custom python environment!'
+            msg = ('No supported QuickJS package found on current python environment!'
+                   ' Please install python package quickjs')
             if _d.chakra_available:
-                msg += ' Please install python package quickjs or use ChakraJSEngine.'
+                msg += ' or use ChakraJSEngine.'
+            elif _d.v8_available:
+                msg += ' or use V8JSEngine.'
             elif _d.external_interpreter:
-                msg += ' Please install python package quickjs or use ExternalJSEngine.'
+                msg += ' or use ExternalJSEngine.'
             else:
-                msg += ' Please install python package quickjs.'
+                msg += '.'
             raise RuntimeError(msg)
         InternalJSEngine.__init__(self, source, init_global, init_del_gobjects)
 
@@ -76,7 +124,9 @@ class QuickJSEngine(InternalJSEngine):
     class Context(object):
         def __init__(self):
             self._context = _d.quickjs.Context()
-            self.typeof = self.Function(self, self._context.eval(u'(obj => typeof obj)'))
+            self.typeof = self._context.eval(u'(obj => typeof obj)')
+            if hasattr(self, 'Function'):
+                self.typeof = self.Function(self.typeof)
 
         def eval(self, code, eval=True, raw=False):
             try:
@@ -88,16 +138,21 @@ class QuickJSEngine(InternalJSEngine):
                     if raw or not isinstance(result, _d.quickjs.Object):
                         return result
                     elif callable(result) and self.typeof(result) == u'function':
-                        return self.Function(self, result)
+                        if hasattr(self, 'Function'):
+                            result = self.Function(result)
+                        return result
                     else:
                         return json.loads(result.json())
 
-        class Function(object):
-            # PetterS/quickjs/Issue7
-            # Escape StackOverflow when calling function outside
-            def __init__(self, context, function):
-                self._context = context
-                self._function = function
+        if _d.quickjs_available and \
+                not hasattr(_d.quickjs.Function, 'execute_pending_job'):
+            # this is < v1.17.0
+            # It was fixed in v1.16.0, but there is no version string to judge.
+            class Function(object):
+                # https://github.com/PetterS/quickjs/issues/7
+                # Escape StackOverflow when calling function outside
+                def __init__(self, function):
+                    self._function = function
 
-            def __call__(self, *args):
-                return self._function(*args)
+                def __call__(self, *args):
+                    return self._function(*args)
